"""Main handler for supr and ldap connections."""

from textwrap import dedent
import ldap
import ldap.modlist as modlist
import sys
import supr
import settings
import subprocess

from requests import ConnectionError
# requests.packages.urllib3.disable_warnings()

from supr_common import (asciify,
                         sendMail,
                         setup_log,)


class SUPR_LDAP:
    """Handle the connection to supr and ldap. (?)"""

    def __init__(self, all_projects, persons_modified, supr_connection):

        if all_projects or persons_modified:

            # Get Logging File Handler
            self.logger = setup_log(self.__class__.__name__, settings.LOG_FILE)

            # LDAP connection
            self.l = None # TODO: Give this variable a better name

            # Get supr_connection
            self.supr_connection    = supr_connection

            # LDAP search scope and other variables
            self.searchScope        = ldap.SCOPE_SUBTREE
            self.retrievedAttrs     = None

            # Variables needed for calculating number of projects and
            # persons created that are used to sendMail to Admins
            self.add_proj_cnt       = 0
            self.mod_proj_cnt       = 0
            self.err_proj_cnt       = 0

            self.add_pers_cnt       = 0
            self.mod_pers_cnt       = 0
            self.err_pers_cnt       = 0
            self.sua_pers_cnt       = 0

            self.krb_pers_cnt       = 0

            self.err_mjr_cnt        = 0
            self.msgtext            = ""

            # Variables needed for Message Text of projects and persons created
            # that are used to sendMail to Admins
            self.ADD_PERS_MAIL  = dedent(f"""\
            The Following Persons have been Added on {settings.today}
            ----------------------------------------------------
            """)

            self.MOD_PERS_MAIL  = "The Following Persons have been Modified on " + settings.today + "\n"
            self.MOD_PERS_MAIL += "---------------------------------------------------- \n"

            self.SUA_PERS_MAIL  = "The Following Persons have not signed SUA and hence not Added on " + settings.today + "\n"
            self.SUA_PERS_MAIL += "---------------------------------------------------------------------------- \n"

            self.KRB_PERS_MAIL  = "The Following Persons have been Added to Kerberos " + settings.today + "\n"
            self.KRB_PERS_MAIL += "------------------------------------------------------- \n"

            self.ADD_PROJ_MAIL  = "The Following Projects have been Added on " + settings.today + "\n"
            self.ADD_PROJ_MAIL += "----------------------------------------------------- \n"

            self.MOD_PROJ_MAIL  = "The Following Projects have been Modified on " + settings.today + "\n"
            self.MOD_PROJ_MAIL += "----------------------------------------------------- \n"

            self.ERR_PROJ_MAIL  = "Error in creating the following Projects on " + settings.today + "\n"
            self.ERR_PROJ_MAIL += "------------------------------------------------------- \n"

            self.ERR_PERS_MAIL  = "Error in creating the following Persons on " + settings.today + "\n"
            self.ERR_PERS_MAIL += "------------------------------------------------------ \n"

            self.ERR_MJR_MAIL   = "Major Error Occured on " + settings.today + "\n"
            self.ERR_MJR_MAIL  += "---------------------------------- \n"

            # Connect to LDAP using python-LDAP
            try:
                ldap.set_option(ldap.OPT_X_TLS_CACERTFILE,
                                settings.TLS_CACERTFILE)
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT,
                                ldap.OPT_X_TLS_NEVER)

                self.l = ldap.initialize(settings.LDAP_HOST)
                self.l.simple_bind(settings.LDAP_ADMIN, settings.LDAP_PASSWORD)
            except ldap.LDAPError as le:
                self.logger.error("LDAP Connection Error - %s", le)
                if self.l:
                    self.l.unbind()

                self.ERR_MJR_MAIL += "Error occured in LDAP Connection - " + str(le) + "\n"
                self.err_mjr_cnt  += 1
                sys.exit(1)

            # Call to addUpdateProjects() Function
            if all_projects:
                # Assigning the all_projects from supr_swestore_main
                self.all_projects = all_projects
                self.addUpdateProjects()

            # Call to updateDeleteProjects() Function
            if persons_modified:
                # Assigning the persons_modified from supr_swestore_main
                self.persons_modified = persons_modified
                self.updateDeletePersons()

            # Call to sendMail() Function
            if self.err_mjr_cnt:
                self.msgtext += str(self.ERR_MJR_MAIL)  + "\n"
            if self.err_proj_cnt:
                self.msgtext += str(self.ERR_PROJ_MAIL) + "\n"
            if self.err_pers_cnt:
                self.msgtext += str(self.ERR_PERS_MAIL) + "\n"
            if self.add_proj_cnt:
                self.msgtext += str(self.ADD_PROJ_MAIL) + "\n"
            if self.mod_proj_cnt:
                self.msgtext += str(self.MOD_PROJ_MAIL) + "\n"
            if self.add_pers_cnt:
                self.msgtext += str(self.ADD_PERS_MAIL) + "\n"
            if self.sua_pers_cnt:
                self.msgtext += str(self.SUA_PERS_MAIL) + "\n"

            if self.krb_pers_cnt:
                self.msgtext += str(self.KRB_PERS_MAIL) + "\n"

            if self.mod_pers_cnt:
                self.msgtext += str(self.MOD_PERS_MAIL) + "\n"
            if self.msgtext:
                sendMail(self.msgtext,
                         settings.FROM_ADDRS,
                         settings.DCACHE_ADMIN_MAIL,
                         settings.LDAP_SUBJECT)

    def sendUserMail(self, m, proj_name, sua_accepted):
        """Mail to be to be sent to User"""

        msgTxt  = "Dear " + m.first_name + " " + m.last_name + ",\n \n"
        msgTxt += "You are receiving this mail as you are added as a member to Swestore project -- " + proj_name + " in SUPR.\n\n"

        if not sua_accepted:
            msgTxt += "We have noticed that you have not signed the NAISS User Agreement.\n"
            msgTxt += "Before access to Swestore is granted, the NAISS User Agreement must be read and accepted.\n"
            msgTxt += "Login to SUPR (https://supr.naiss.se/login), go to Personal Information and sign the NAISS User Agreement using Federated Identity. \n\n"

        msgTxt += "For information on how to set your password, please see the following documentation :-\n"
        msgTxt += "https://docs.swestore.se/access/\n\n"

        msgTxt += "Please refer to the documentation for more information on Swestore:-\n"
        msgTxt += "https://docs.swestore.se/\n\n"

        msgTxt += "In case of any issues or clarifications, please use the support form on\n"
        msgTxt += "https://supr.naiss.se/support/ or email support@swestore.se.\n\n"
        msgTxt += "Regards,\nSwestore Support Team\n"

        sendMail(msgTxt,
                 'support@swestore.se',
                 m.email,
                 "Information to access Swestore")

    def getUID(self, m):
        """Generate userId or uid for new Person from first name and last
        name using various combinations"""

        guesses   = []
        firstname = ((m.first_name).replace(" ", "")).lower()
        lastname  = (m.last_name).replace(" ", "").lower()

        if (len(firstname) < 3 and len(lastname) < 3):
            uid = "s_" + asciify(u'' + firstname + lastname)
            guesses.append(str(uid))
        else:
            for (flen, llen, swap) in ((3, 2, False),
                                       (2, 3, False),
                                       (3, 2, True),
                                       (2, 3, True),
                                       (1, 4, False),
                                       (4, 1, False),
                                       (1, 4, True),
                                       (4, 1, True),
                                       (0, 5, False),
                                       (5, 0, False)):
                if (len(firstname) < flen or len(lastname) < llen):
                    continue
                if swap:
                    uid = "s_" + asciify(u'' + lastname[:llen] + firstname[:flen])
                    guesses.append(str(uid))
                else:
                    uid = "s_" + asciify(u'' + firstname[:flen] + lastname[:llen])
                    guesses.append(str(uid))

        # Check if the uid already exists in LDAP
        self.logger.debug("Person with SUPR ID :: %s login guesses -- %s ",
                          m.id,
                          guesses)

        for login in guesses:

            filter = "uid=" + login
            ldap_result_id = self.l.search(settings.peopleDN,
                                           self.searchScope,
                                           filter,
                                           self.retrievedAttrs)
            result_type, result_data = self.l.result(ldap_result_id, 0)

            if (not result_data):
                return login

        # No guess possible
        return str(guesses[0] + m.id)

    def searchPerson(self, uidNumber):
        """Search Person based on uidNumber """
        try:
            # Check if the member already exists in LDAP
            filter = "(uidNumber=" + uidNumber + ")"

            ldap_result_id = self.l.search(settings.peopleDN,
                                           self.searchScope,
                                           filter,
                                           self.retrievedAttrs)
            result_type, result_data = self.l.result(ldap_result_id, 0)
            self.logger.info("searchPerson -- " + str(result_type) + "--- " + str(result_data))
            return result_data
        except ldap.LDAPError as le:
            self.logger.error("LDAP Error in searchPerson Module for %s :: %s", str(uidNumber), le)
            self.ERR_PERS_MAIL += "uidNumber :: " + str(uidNumber) + "\t Module :: searchPerson \n"
            self.err_pers_cnt  += 1
            return []

    def searchMemberUid(self, uidNumber):
        """Search MemberUid based on uidNumber in Groups for SUP not signed
        persons"""

        try:
            # Check if the member already exists in LDAP
            filter = "(memberUid=" + uidNumber + ")"
            retrievedAttrs = ['gidNumber', 'resourceID', 'cn']

            ldap_result_id = self.l.search(settings.groupsDN, self.searchScope, filter, retrievedAttrs)
            result_type, result_data = self.l.result(ldap_result_id)
            self.logger.info("chk 111 -- " + str(result_type) + " --- " + str(result_data))
            return result_data
        except ldap.LDAPError as le:
            self.logger.error("LDAP Error in searchMemberUid Module for %s :: %s", str(uidNumber), le)
            self.ERR_PERS_MAIL += "uidNumber :: " + str(uidNumber) + "\t Module :: searchMemberUid \n"
            self.err_pers_cnt  += 1
            return []

    def createPersonAttrs(self, m, uidNumber):
        """Return a Person Attributes dictionary"""

        attrsPerson                     = {}
        attrsPerson['objectclass']      = [('Swestore').encode('utf-8'), ('inetorgperson').encode('utf-8'), ('posixaccount').encode('utf-8')]
        attrsPerson['description']      = str(m.id).encode('utf-8')
        attrsPerson['uidNumber']        = uidNumber
        attrsPerson['gidNumber']        = uidNumber
        attrsPerson['gn']               = (m.first_name).encode('utf-8')
        attrsPerson['sn']               = (m.last_name).encode('utf-8')
        attrsPerson['cn']               = (m.first_name + " " + m.last_name).encode('utf-8')
        attrsPerson['mail']             = (m.email).encode('utf-8')
        attrsPerson['subject']          = (m.subject).encode('utf-8')
        attrsPerson['lastModifiedTime'] = (m.modified).encode('utf-8')
        attrsPerson['homeDirectory']    = ("/").encode('utf-8')

        # Will be updated to the correct yubikey value once yubikey is
        # integrated into SUPR (Not needed as of now)
        attrsPerson['yubikey'] = ("yubikey").encode('utf-8')

        # Need to check if centre_person_id comes from SUPR. If it clashes
        # with other UIDs then an alert mail needs to be sent to Admin about
        # the same.
        if m.centre_person_id:
            attrsPerson['uid'] = str(m.centre_person_id).encode('utf-8')
        else:
            attrsPerson['uid'] = str(self.getUID(m)).encode('utf-8')

        #attrsPerson = {key:[v.encode("utf-8") if type(v) == str else v for v in values] for key, values in attrsPerson.items()}

        return attrsPerson

    def personChanged(self, result_data, attrsPerson):
        """Check if Person Data has changed"""
        if ((result_data[0][1].get('description', False)
             and result_data[0][1].get('description', '')[0] != attrsPerson['description'])):
            return True
        if ((result_data[0][1].get('givenName', False)
             and result_data[0][1].get('givenName', '')[0] != attrsPerson['gn'])):
            return True
        if ((result_data[0][1].get('sn', False)
             and result_data[0][1].get('sn', '')[0] != attrsPerson['sn'])):
            return True
        if ((result_data[0][1].get('cn', False)
             and result_data[0][1].get('cn', '')[0] != attrsPerson['cn'])):
            return True
        if ((result_data[0][1].get('mail', False)
             and result_data[0][1].get('mail', '')[0] != attrsPerson['mail'])):
            return True
        if ((result_data[0][1].get('subject') is None and attrsPerson['subject'])
            or (result_data[0][1].get('subject') and result_data[0][1].get('subject', '')[0] != attrsPerson['subject'])):
            return True
        if ((result_data[0][1].get('uid', False)
             and result_data[0][1].get('uid', '')[0] != attrsPerson['uid'])):
            return True

        return False

    def addPerson(self, m, uidNumber, resourceIDList):
        """Add person (to ldap?)"""

        try:
            attrsPerson = self.createPersonAttrs(m, uidNumber)
            self.logger.info("chking addPerson attrsPerson -- " + str(attrsPerson))
            #personDN = "uidNumber=" + uidNumber + "," + peopleDN
            personDN = "uid=" + attrsPerson['uid'] + "," + settings.peopleDN
            self.logger.info("chking addPerson personDN -- " + str(personDN))

            ldif = modlist.addModlist(attrsPerson)
            self.l.add_s(personDN, ldif)

            self.logger.info("\n Person with SUPR ID SUP is signed :: %s and added to LDAP -- %s", m.id, str(uidNumber))

            if settings.dcache_resource_id in resourceIDList:
                self.ADD_PERS_MAIL += "SUPR ID :: " + str(m.id) + "\t username(uid) :: " + attrsPerson['uid'] + "\t Person Name :: " + attrsPerson['cn'] + "\n"
                self.add_pers_cnt  += 1

            if m.centre_person_id:
                self.logger.info("Person with SUPR ID :: %s centre_person_id already exists -- %s", m.id, m.centre_person_id)
            else:
                # Updating the centre_person_id to SUPR
                d = {'centre_person_id': attrsPerson['uid']}

                try:
                    person = self.supr_connection.post("person/%s/update/", m.id, data=d)
                    m.centre_person_id = attrsPerson['uid']
                    self.logger.info("Centre_Person_id for Person with SUPR ID :: %s updated to SUPR -- %s", m.id, attrsPerson['uid'])
                except supr.SUPRHTTPError as e:
                    # We want to show the text received if we get an HTTP Error
                    self.logger.error("HTTP error in updating data to SUPR - %s :: %s" % e.status_code      % e.text)
                    self.ERR_MJR_MAIL += "Error occured in updating centre_person_id (uid) to SUPR - " + e.text + "\n"
                    self.ERR_MJR_MAIL += "SUPR ID :: " + str(m.id) + "\t username(uid) :: " + attrsPerson['uid'] + "\t Person Name :: " + attrsPerson['cn'] + "\t Module :: addPerson \n"
                    self.err_mjr_cnt  += 1
                except ConnectionError as ce:
                    self.ERR_MJR_MAIL += "Error occured in updating centre_person_id (uid) to SUPR due to ConnectionError- " + str(ce) + "\n"
                    self.logger.error("Connection to SUPR failed - %s", str(ce))

            for resourceid in resourceIDList:
                d = {'username': m.centre_person_id,
                     'person_id': m.id,
                     'resource_id': resourceid,
                     'status': 'enabled',
                     'button_text': 'Change password'}
                try:
                    account = self.supr_connection.post("account/create/", data=d)
                    self.logger.info("Account for Person with SUPR ID :: %s for resource -- %s created", m.id, resourceid)
                except supr.SUPRHTTPError as e:
                    # We want to show the text received if we get an HTTP Error
                    self.logger.error("HTTP error in creating account to SUPR - %s :: %s", e.status_code, e.text)
                    self.ERR_MJR_MAIL += "Error occured in creating account to SUPR - " + e.text + "\n"
                    self.ERR_MJR_MAIL += "SUPR ID :: " + str(m.id) + "\t username(uid) :: " + str(attrsPerson['uid']) + "\t Module :: addPerson \n"
                    self.err_mjr_cnt  += 1
                except ConnectionError as ce:
                    self.ERR_MJR_MAIL += "Error occured in creating account to SUPR due to ConnectionError- " + str(ce) + "\n"
                    self.logger.error("Connection to SUPR failed - %s", str(ce))

            # Adding Project for every Person
            self.addPersonAsProject(attrsPerson)

            # Adding Person to Kerberos
            self.addPersontoKerberos(m, attrsPerson)

        except ldap.LDAPError as le:
            self.logger.error("LDAP Error in addPerson Module for %s :: %s", str(uidNumber), le)
            self.ERR_PERS_MAIL += "uidNumber :: " + str(uidNumber) + "\t Module :: addPerson \n"
            self.err_pers_cnt  += 1

    def addPersontoKerberos(self, m, attrsPerson):
        """Add a new person to Kerberos, if able

        Uses external `kadmin` program.

        :param m: Person to add (?)
        :type m: Ldap person object (?)
        :param attrsPerson: Person attributes to add to kerberos (?)
        :type attrsPerson: Dictionary
        :returns: None
        """

        try:
            # TODO: This one needs testing
            result = subprocess.run([
                "kadmin",
                "-k", "-t", settings.keytab,
                "-p", settings.adminprincipal,
                "addprinc", "-randkey", attrsPerson['uid']],

                timeout=10, text=True, capture_output=True)

            for line in result.stdout.splitlines():
                self.logger.info(line)
            for line in result.stderr.splitlines():
                self.logger.warning(line)

            if result.returncode == 0:
                self.logger.info("Person with SUPR ID :: %s added to Kerberos -- %s",
                                 m.id, str(attrsPerson['uidNumber']))
                self.KRB_PERS_MAIL += "SUPR ID :: " + str(m.id) + "\t username(uid) :: " + attrsPerson['uid'] + "\t Person Name :: " + attrsPerson['cn'] + "\n"
                self.krb_pers_cnt  += 1
            else:
                self.logger.info("Person with SUPR ID :: %s already added to Kerberos -- %s", m.id, str(attrsPerson['uidNumber']))
                sys.exit(1)

        except Exception as exc:
            self.logger.error("Error in addPersontoKerberos Module for %s :: %s", str(attrsPerson['uid']), exc)
            self.ERR_PERS_MAIL += "uidNumber :: " + str(attrsPerson['uid']) + "\t Module :: addPersontoKerberos \n"
            self.err_pers_cnt  += 1

    def updateDeletePersons(self):
        """Update Person Data or delete merged Persons"""

        self.logger.info("updateDeletePersons -- " + str(self.persons_modified))
        gidNumbers = []
        try:
            for m in self.persons_modified:

                uidNumber       = str(settings.uidNumberStart + m.id)
                result_data     = self.searchPerson(uidNumber)
                self.logger.info("updateDeletePersons searchPerson result_data -- " + str(result_data))

                if ((not result_data) and (not m.merged_ids)):

                    # check for uidnumber in memberUids in groups
                    gidNumbers = self.searchMemberUid(uidNumber)
                    self.logger.info("gidNumbers -- " + str(gidNumbers))

                    if gidNumbers:
                        if (m.user_agreement_version and m.user_agreement_accepted):

                            for gidNumber in gidNumbers:

                                gid = gidNumber[1].get('gidNumber')[0]
                                resourceIDList = gidNumber[1].get('resourceID')
                                self.logger.info("resourceIDList : %s ", resourceIDList)

                                proj_name = gidNumber[1].get('cn')[0]

                                self.addPerson(m, uidNumber, resourceIDList)

                                groupDN = "gidNumber=" + gid + "," + settings.groupsDN
                                oldMemberUid = [(ldap.MOD_DELETE, 'memberUid', str(uidNumber))]
                                newMemberUid = [(ldap.MOD_ADD, 'memberUid', str(m.centre_person_id))]

                                self.l.modify_s(groupDN, oldMemberUid)
                                self.l.modify_s(groupDN, newMemberUid)

                            self.logger.info("Person with SUPR ID :: %s SUP is signed and will be updated to LDAP.", m.id)

                    else:
                        self.logger.info("Person with SUPR ID :: %s not needed in LDAP. Just ignore it.", m.id)
                else:

                    for merge_id in m.merged_ids:
                        mergeUidNumber   = str(settings.uidNumberStart + merge_id)
                        result_data_mrg  = self.searchPerson(mergeUidNumber)

                        if(not result_data_mrg):
                            self.logger.info("Person with SUPR ID :: %s has already been deleted from LDAP since its merged with SUPR ID :: %s ", merge_id, m.id)
                        else:
                            personDN = "uid=" + result_data_mrg[0][1].get('uid')[0] + "," + settings.peopleDN
                            groupDN  = "gidNumber=" + mergeUidNumber + "," + settings.groupsDN

                            self.l.delete_s(personDN)
                            self.logger.info("Person with SUPR ID :: %s deleted from LDAP since its merged with SUPR ID :: %s ", merge_id, m.id)

                            self.MOD_PERS_MAIL += "SUPR Person ID :: " + str(merge_id) + " merged with SUPR Person ID :: " + str(m.id) + "\n"
                            self.mod_pers_cnt  += 1

                            # Deleting the corresponding Project as well
                            self.l.delete_s(groupDN)
                            self.logger.info("Project with ID :: %s deleted from LDAP since its person is also deleted :: %s ", merge_id, merge_id)

                    attrsPerson = self.createPersonAttrs(m, uidNumber)

                    self.logger.info("updateDeletePerson attrsPerson -- " + str(attrsPerson))

                    personDN = "uid=" + str(attrsPerson['uid']) + "," + settings.peopleDN

                    self.logger.info("updateDeletePerson personDN -- " + str(personDN))

                    if(result_data and self.personChanged(result_data, attrsPerson)):
                        self.l.delete_s(personDN)
                        ldif = modlist.addModlist(attrsPerson)
                        self.l.add_s(personDN, ldif)
                        self.logger.info("Person with SUPR ID :: %s modified to LDAP -- %s", m.id, attrsPerson['uidNumber'])

                        self.MOD_PERS_MAIL += "SUPR ID :: " + str(m.id) + "\t username(uid) :: " + attrsPerson['uid'] + "\t Person Name :: " + attrsPerson['cn'] + "\n"
                        self.mod_pers_cnt  += 1

                    else:
                        self.logger.info("Person with SUPR ID :: %s - No changes since last time \n", m.id)

        except ldap.LDAPError as le:
            self.logger.error("LDAP Error in updateDeletePersons Module for %s :: %s", str(uidNumber), le)
            self.ERR_PERS_MAIL += "uidNumber :: " + str(uidNumber) + "\t Module Name :: updateDeletePerson \n"
            self.err_pers_cnt  += 1

    def searchProject(self, gidNumber, cn):
        """Search Project based on gidNumber """
        try:
            # Check if the project already exists in LDAP
            filter = "(|(cn=" + cn + ")(gidNumber=" + gidNumber + "))"

            ldap_result_id = self.l.search(settings.groupsDN, self.searchScope, filter, self.retrievedAttrs)
            result_type, result_data = self.l.result(ldap_result_id)
            self.logger.info("result_data full -- " + str(result_data))

            for result in result_data:
                self.logger.info("result -- " + str(result[1]['cn'][0]))
                if str(cn) == str(result[1]['cn'][0]):
                    return [result]

        except ldap.LDAPError as le:
            self.logger.error("LDAP Error in searchProject Module for %s :: %s", str(gidNumber), le)
            self.ERR_PROJ_MAIL += "gidNumber :: " + str(gidNumber) + "\t Module :: searchProject \n"
            self.err_proj_cnt  += 1
            return []

    def addUpdateProjects(self):
        """Add or Update Projects"""

        for p in self.all_projects:

            attrsGroup      = {}
            memberUIDList   = []
            resourceIDList  = []
            suaNotSigned    = []
            sua_accepted    = None

            for rp in p.resourceprojects:
                if str(rp.resource.id) == settings.dcache_resource_id:
                    resourceIDList.append(str(rp.resource.id))

            for m in p.members:
                # Generate uidNumber for new Person
                # ( Example 30023 - 30000 is the start range and 23 is the person id from SUPR )
                uidNumber = str(settings.uidNumberStart + m.id)
                result_data = self.searchPerson(uidNumber)
                self.logger.info("addUpdateProjects searchPerson result_data -- " + str(result_data))
                result_data_mid = self.searchMemberUid(uidNumber)
                self.logger.info("addUpdateProjects searchPerson result_data_mid -- " + str(result_data_mid))

                if ((not result_data) and (not result_data_mid)):

                    if (m.user_agreement_version and m.user_agreement_accepted):
                        self.logger.info("addUpdateProjects m.id -- " + str(m.id) + " merged_ids -- " + str(m.merged_ids))

                        if m.merged_ids:
                            self.persons_modified = [m]
                            self.updateDeletePersons()

                        self.addPerson(m, uidNumber, resourceIDList)
                        sua_accepted = True

                        # To send user mail regarding user certificate registration in SUPR for dcache projects
                        if settings.dcache_resource_id in resourceIDList:
                            self.sendUserMail(m, p.directory_name, sua_accepted)
                    else:
                        sua_accepted = False
                        suaNotSigned.append(str(m.id))
                        # To send user mail regarding SUA in SUPR for all projects
                        self.sendUserMail(m, p.directory_name, sua_accepted)
                        self.logger.info("Person with SUPR ID :: %s SUP is not signed and will not be added to LDAP.", m.id)

                        # Add UidNumber as MemberUid in Group fo SUP not signed
                        m.centre_person_id = uidNumber

                elif (result_data and result_data[0][1].get('uid')[0]):
                    m.centre_person_id = result_data[0][1].get('uid')[0]

                    self.logger.info("Person with SUPR ID :: %s - Already added to LDAP", m.id)

                elif result_data_mid:
                    m.centre_person_id = uidNumber

                if not m.centre_person_id:
                    m.centre_person_id = uidNumber

                memberUIDList.append(str(m.centre_person_id))

            if (suaNotSigned and (settings.dcache_resource_id in resourceIDList)):
                self.SUA_PERS_MAIL += "SUPR ID :: " + str(p.id) + "\t Project Name :: " + str(p.directory_name) + "\t Persons SUA Not signed :: " + ", ".join(suaNotSigned) + "\n"
                self.sua_pers_cnt  += 1

            # Generate gidNumber for new Project
            # ( Example 6012 - 6000 is the start range and 12 is the project id from SUPR )
            gidNumber = str(settings.gidNumberStart + p.id)

            groupDN                        = "gidNumber=" + gidNumber + "," + settings.groupsDN
            attrsGroup['objectclass']      = [b'Swestore', b'posixgroup']
            attrsGroup['description']      = [str(p.id).encode("utf-8")]
            attrsGroup['piMemberID']       = (str(settings.uidNumberStart + p.pi.id)).encode("utf-8")
            attrsGroup['memberUID']        = [m.encode("utf-8") for m in memberUIDList]
            attrsGroup['resourceID']       = [r.encode("utf-8") for r in resourceIDList]
            attrsGroup['projectEndDate']   = str(p.end_date).encode("utf-8")
            attrsGroup['lastModifiedTime'] = str(p.modified).encode("utf-8")

            if p.directory_name:
                attrsGroup['cn'] = p.directory_name.encode('utf-8')

            try:
                result_data = self.searchProject(gidNumber, str(attrsGroup['cn']))
                self.logger.info("attrsGroup 1-- %s ", str(attrsGroup))
                self.logger.info("result_data 1-- %s ", str(result_data))

                # Load data into LDAP
                if (not result_data):
                    ldif = modlist.addModlist(attrsGroup)
                    self.l.add_s(groupDN, ldif)
                    self.logger.info("Project with SUPR ID :: %s added to LDAP -- %s \n", p.id, gidNumber)

                    if settings.dcache_resource_id in resourceIDList:
                        self.ADD_PROJ_MAIL += "SUPR ID :: " + str(p.id) + "\t Project Name :: " + p.directory_name + "\t Persons Added :: " + ", ".join(memberUIDList) + "\n"
                        self.add_proj_cnt  += 1

                else:
                    if (result_data
                        and ('lastModifiedTime' in result_data[0][1])
                        and (result_data[0][1]['lastModifiedTime'][0] != attrsGroup['lastModifiedTime'])):

                        addUsers        = list(set(memberUIDList).difference(set(result_data[0][1]['memberUid'])))
                        removeUsers     = list(set(result_data[0][1]['memberUid']).difference(set(memberUIDList)))
                        suprIds         = result_data[0][1]['description']

                        if p.continuation_name:
                            gidNumber = result_data[0][1]['gidNumber'][0]
                            groupDN   = "gidNumber=" + gidNumber + "," + settings.groupsDN
                            if str(p.id) not in suprIds:
                                attrsGroup['description']  = suprIds + [str(p.id)]
                            else:
                                attrsGroup['description']  = suprIds

                        self.logger.info("attrsGroup 2 -- %s ", str(attrsGroup))
                        self.l.delete_s(groupDN)
                        ldif = modlist.addModlist(attrsGroup)
                        self.l.add_s(groupDN, ldif)

                        self.logger.info("Project with SUPR ID :: %s modified to LDAP -- %s ", p.id, gidNumber)

                        if p.continuation_name:
                            self.MOD_PROJ_MAIL += "SUPR ID :: " + str(p.id) + "\t Project Name :: " + attrsGroup['cn'] + "\t Continuation project for SUPR ID(s) :: " + ", ".join(suprIds) + "\n"
                            self.mod_proj_cnt  += 1

                        newCn = False
                        if attrsGroup['cn'] != result_data[0][1]['cn'][0]:
                            self.MOD_PROJ_MAIL += "SUPR ID :: " + str(p.id) + "\t New Project Name :: " + attrsGroup['cn'] + "\t Old Project Name :: " + result_data[0][1]['cn'][0]  + "\n"
                            self.mod_proj_cnt  += 1
                            newCn = True

                        if removeUsers and (settings.dcache_resource_id in resourceIDList):
                            self.MOD_PROJ_MAIL += "SUPR ID :: " + str(p.id) + "\t Project Name :: " + attrsGroup['cn'] + "\t Persons Removed :: " + ", ".join(removeUsers) + "\n"
                            self.mod_proj_cnt  += 1

                        if addUsers and (settings.dcache_resource_id in resourceIDList):
                            self.MOD_PROJ_MAIL += "SUPR ID :: " + str(p.id) + "\t Project Name :: " + attrsGroup['cn'] + "\t Persons Added   :: " + ", ".join(addUsers) + "\n"
                            self.mod_proj_cnt  += 1

                        if not (removeUsers or addUsers or p.continuation_name or newCn):
                            self.MOD_PROJ_MAIL += "SUPR ID :: " + str(p.id) + "\t Project Name :: "  + attrsGroup['cn'] + "\n"
                            self.mod_proj_cnt  += 1

                    else:
                        self.logger.info("Project with SUPR ID :: %s - No changes since last time \n", p.id)

            except ldap.LDAPError as le:
                self.logger.error("LDAP Error in addUpdateProjects Module for %s :: %s", str(gidNumber), le)
                self.ERR_PROJ_MAIL += "gidNumber :: " + str(gidNumber) + "\t Module :: addUpdateProjects \n"
                self.err_proj_cnt  += 1

    def addPersonAsProject(self, attrsPerson):
        """For every Person add a corresponding Project with uidNumber as
        gidNumber"""

        attrsGroup = {}
        attrsGroup['cn'] = attrsPerson['uid']
        attrsGroup['description'] = attrsPerson['description']
        attrsGroup['piMemberID'] = attrsPerson['uidNumber']
        attrsGroup['memberUID'] = [attrsPerson['uid']]
        attrsGroup['objectclass'] = ['Swestore', 'posixgroup']

        gidNumber = attrsPerson['uidNumber']
        attrsGroup['gidNumber'] = gidNumber
        groupDN = "gidNumber=" + gidNumber + "," + settings.groupsDN

        try:
            result_data = self.searchProject(gidNumber, attrsGroup['cn'])

            # Load data into LDAP
            if (not result_data):
                ldif = modlist.addModlist(attrsGroup)
                self.l.add_s(groupDN, ldif)
                self.logger.info("Person-Project with ID :: %s added to LDAP -- %s", attrsPerson['description'], attrsGroup['gidNumber'])

            else:
                self.logger.info("Person-Project with ID :: %s - already added", attrsPerson['description'])

        except ldap.LDAPError as le:
            self.logger.error("LDAP Error in addPersonAsProject Module for %s :: %s", str(gidNumber), le)
            self.ERR_PROJ_MAIL += "gidNumber :: " + str(gidNumber) + "\t Module :: addPersonAsProject \n"
            self.err_proj_cnt  += 1
